"""Utils."""
from typing import Callable


def cache_decorator(size: int=512) -> Callable:
    """Dummy cache decorator."""
    _cache = {}

    def decorator(func):
        def _wrapper(key):
            if key in _cache:
                return _cache[key]

            _cache[key] = res = func(key)

            if len(_cache) > size:
                _cache.pop(next(iter(_cache)))
            return res

        _wrapper.cache = _cache
        return _wrapper

    return decorator



"""Connector stuffs."""

from asyncio import wait_for
from asyncio import sleep as asyncio_sleep
from ssl import SSLContext
from typing import Coroutine
from urllib.parse import ParseResult

# import h2.connection (unused)
from hyperframe.frame import SettingsFrame

# from concurrent import futures (unused)
from aiosonic.exceptions import ConnectionPoolAcquireTimeout
from aiosonic.exceptions import TimeoutException
from aiosonic.pools import SmartPool
from aiosonic.timeout import Timeouts


class TCPConnector:
    """TCPConnector.

    Holds the main logic for making connections to destination hosts.

    Params:
        * **pool_size**: size for pool of connections
        * **timeouts**: global timeouts to use for connections with this connector. default: :class:`aiosonic.timeout.Timeouts` instance with default args.
        * **connection_cls**: connection class to be used. default: :class:`aiosonic.connection.Connection`
        * **pool_cls**: pool class to be used. default: :class:`aiosonic.pools.SmartPool`

    """

    def __init__(self,
                 pool_size: int = 25,
                 timeouts: Timeouts = None,
                 connection_cls=None,
                 pool_cls=None):
        from aiosonic.connection import Connection  # avoid circular dependency
        self.pool_size = pool_size
        connection_cls = connection_cls or Connection
        pool_cls = pool_cls or SmartPool
        self.pool = pool_cls(self, pool_size, connection_cls)
        self.timeouts = timeouts or Timeouts()

    async def acquire(self, urlparsed: ParseResult):
        """Acquire connection."""
        # Faster without timeout
        if not self.timeouts.pool_acquire:
            return await self.pool.acquire(urlparsed)

        try:
            return await wait_for(self.pool.acquire(urlparsed),
                                  self.timeouts.pool_acquire)
        except TimeoutException:
            raise ConnectionPoolAcquireTimeout()

    async def release(self, conn):
        """Release connection."""
        res = self.pool.release(conn)
        if isinstance(res, Coroutine):
            await res

    async def wait_free_pool(self):
        """Wait until free pool."""
        while True:
            if self.pool.is_all_free():
                return True
            asyncio_sleep(0.02)  # pragma: no cover

    async def cleanup(self):
        """Cleanup connector connections."""
        await self.pool.cleanup()



from typing import Optional


class Timeouts():
    """Timeouts class wrapper.

    Arguments:
        * sock_connect(float): time for establish connection to server
        * sock_read(float): time until get first read
        * pool_acquire(float): time until get connection from
          connection's pool
        * request_timeout(float): time until complete request.
    """

    def __init__(self,
                 sock_connect: Optional[float] = 5,
                 sock_read: Optional[float] = 30,
                 pool_acquire: Optional[float] = None,
                 request_timeout: Optional[float] = 60) -> None:
        self.sock_connect = sock_connect
        self.sock_read = sock_read
        self.pool_acquire = pool_acquire
        self.request_timeout = request_timeout



"""Pools module."""

from urllib.parse import ParseResult
from asyncio import Semaphore
from asyncio import Queue


class CyclicQueuePool:
    """Cyclic queue pool of connections."""
    def __init__(self, connector, pool_size, connection_cls):
        self.pool_size = pool_size
        self.pool = Queue(pool_size)

        for _ in range(pool_size):
            self.pool.put_nowait(connection_cls(connector))

    async def acquire(self, _urlparsed: ParseResult = None):
        """Acquire connection."""
        return await self.pool.get()

    async def release(self, conn):
        """Release connection."""
        return self.pool.put_nowait(conn)

    def is_all_free(self):
        """Indicates if all pool is free."""
        return self.pool_size == self.pool.qsize()

    def free_conns(self) -> int:
        return self.pool.qsize()

    async def cleanup(self):
        """Get all conn and close them, this method let this pool unusable."""
        for _ in range(self.pool_size):
            conn = self.pool.get()
            conn.close()


class SmartPool:
    """Pool which utilizes alive connections."""
    def __init__(self, connector, pool_size, connection_cls):
        self.pool_size = pool_size
        self.pool = set()
        self.sem = Semaphore(pool_size)

        for _ in range(pool_size):
            self.pool.add(connection_cls(connector))

    async def acquire(self, urlparsed: ParseResult = None):
        """Acquire connection."""
        await self.sem.acquire()
        if urlparsed:
            key = f'{urlparsed.hostname}-{urlparsed.port}'
            for item in self.pool:
                if item.key == key:
                    self.pool.remove(item)
                    return item
        return self.pool.pop()

    def release(self, conn) -> None:
        """Release connection."""
        self.pool.add(conn)
        self.sem.release()

    def free_conns(self) -> int:
        return len(self.pool)

    def is_all_free(self):
        """Indicates if all pool is free."""
        return self.pool_size == self.sem._value

    async def cleanup(self) -> None:
        """Get all conn and close them, this method let this pool unusable."""
        for count in range(self.pool_size):
            conn = await self.acquire()
            conn.close()



from typing import Union
from typing import Dict
from typing import Sequence
from typing import Tuple
from typing import AsyncIterator
from typing import Iterator

# TYPES
ParamsType = Union[Dict[str, str], Sequence[Tuple[str, str]], ]
#: Data to be sent in requests, allowed types
DataType = Union[str, bytes, dict, tuple, AsyncIterator[bytes],
                 Iterator[bytes], ]
BodyType = Union[str, bytes, AsyncIterator[bytes], Iterator[bytes], ]
ParsedBodyType = Union[bytes, AsyncIterator[bytes], Iterator[bytes], ]



"""Setup module."""

import re
from setuptools import setup


RGX = re.compile(r'([\w-]+[<>=]{1}=[\d.\w]+)')


def read_file(filename):
    """Read file correctly."""
    with open(filename) as _file:
        return _file.read().strip()


def requirements(filename):
    """Parse requirements from file."""
    return re.findall(RGX, read_file(filename)) or []


def version():
    data = read_file('./aiosonic/version.py')
    return re.findall(r"VERSION = '([a-z0-9.]*)'", data)[0]


# copied form uvicorn, mark to not install uvloop in windows
env_marker = (
    "sys_platform != 'win32'"
    " and sys_platform != 'cygwin'"
    " and platform_python_implementation != 'PyPy'"
)

pypy_marker = (
    "platform_python_implementation != 'PyPy'"
)


def add_marks(dependencies, marks):
    """Add markers to dependencies.

    Example:
        uvloop==0.12.0 -> uvloop==0.12.0 ; sys_platform != 'win32'...
    """
    def _map_func(dependency):
        for item, marker in marks.items():
            if item in dependency:
                return dependency + marker
        return dependency

    return list(map(_map_func, dependencies))


setup(
    name='aiosonic',
    version=version(),
    description='Async http client',
    author='Johanderson Mogollon',
    author_email='johanderson@mogollon.com.ve',
    long_description=read_file('README.md'),
    long_description_content_type='text/markdown',
    license='MIT',
    packages=[
        'aiosonic',
        'aiosonic_utils',
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Internet :: WWW/HTTP',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    setup_requires=['pytest-runner'],
    install_requires=requirements('./requirements.txt'),
    extras_require={
        'test': add_marks(
            requirements('./test-requirements.txt'),
            {
                'uvloop': ' ;' + env_marker,
                'httptools': ' ;' + env_marker,
                'mypy': ' ;' + pypy_marker,
                'mypy-extensions': ' ;' + pypy_marker,
                'pytest-mypy': ' ;' + pypy_marker,
                'typed-ast': ' ;' + pypy_marker,
            }
        )
    }
)
